#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <iostream>
#include <ctime>
#include <cmath>

#include <amdis/AMDiS.hpp>
#include <amdis/LocalOperators.hpp>
#include <amdis/ProblemStat.hpp>

#ifdef DOW
#undef DOW
#endif
#define DOW WORLDDIM

using namespace AMDiS;

// 3 components: velocity with polynomial degree 2 and pressure with polynomial degree 1

using Grid = Dune::YaspGrid<GRIDDIM>;
using StokesParam   = TaylorHoodBasis<Grid::LeafGridView>;
using StokesProblem = ProblemStat<StokesParam>;

int main(int argc, char** argv)
{
  AMDiS::init(argc, argv);

  StokesProblem prob("stokes");
  prob.initialize(INIT_ALL);

  double viscosity = 1.0;
  Parameters::get("stokes->viscosity", viscosity);

  // tree-paths for components
  auto _v = Dune::Indices::_0;
  auto _p = Dune::Indices::_1;

  // <viscosity*grad(u_i), grad(v_i)>
  for (std::size_t i = 0; i < DOW; ++i) {
    auto opL = makeOperator(tag::gradtest_gradtrial{}, viscosity);
    prob.addMatrixOperator(opL, treepath(_v,i), treepath(_v,i));

    // <d_i(v_i), p>
    auto opP = makeOperator(tag::partialtest_trial{i}, 1.0);
    prob.addMatrixOperator(opP, treepath(_v,i), _p);

    // <q, d_i(u_i)>
    auto opDiv = makeOperator(tag::test_partialtrial{i}, 1.0);
    prob.addMatrixOperator(opDiv, _p, treepath(_v,i));
  }

  auto opZero = makeOperator(tag::test_trial{}, 0.0);
  prob.addMatrixOperator(opZero, _p, _p);

  // define boundary regions
  auto left = [](auto const& x) { return x[0] < 1.e-8; };
  auto not_left = [](auto const& x) { return x[0] > 1.0 - 1.e-8 || x[1] < 1.e-8 || x[1] > 1.0 - 1.e-8; };

  // define boundary values
  auto parabolic_y = [](auto const& x) -> Dune::FieldVector<double,DOW>
  {
    return {0.0, x[1]*(1.0 - x[1])};
  };

  auto zero = [](auto const& x) -> Dune::FieldVector<double,DOW>
  {
    return {0.0, 0.0};
  };

  // set boundary conditions for velocity
  prob.addDirichletBC(left, _v, _v, parabolic_y);
  prob.addDirichletBC(not_left, _v, _v, zero);

  prob.addDirichletBC([](auto const& x) { return x[0] < 1.e-8 && x[1] < 1.e-8; }, _p, _p, 0.0);

  AdaptInfo adaptInfo("adapt");

  // assemble and solve system
  prob.assemble(adaptInfo);

#ifdef DEBUG_MTL
  // write matrix to file
  mtl::io::matrix_market_ostream out("matrix_stokes0.mtx");
  out << prob.systemMatrix().matrix();
  std::cout << prob.systemMatrix().matrix() << '\n';
#endif

  prob.solve(adaptInfo);

  // output solution
  prob.writeFiles(adaptInfo);

  AMDiS::finalize();
  return 0;
}
