#pragma once

#include <list>

#include <amdis/Output.hpp>
#include <amdis/linearalgebra/DOFVectorInterface.hpp>

namespace AMDiS
{
  /**
   * \addtogroup Adaption
   * @{
   **/

  /// \brief Interface for transfer between grid changes to be registered in a \ref GridTransferManager
  class GridTransferInterface
  {
  public:
    virtual ~GridTransferInterface() = default;

    /// Attach a data container to the grid transfer, that gets interpolated during grid change
    virtual void attach(DOFVectorInterface*) = 0;

    /// Remove a data constainer from the grid transfer. Throws a warning if container not found.
    virtual void detach(DOFVectorInterface*) = 0;

    /// Prepare the grid and the data for the adaption
    virtual bool preAdapt() = 0;

    /// Do the grid adaption
    virtual bool adapt() = 0;

    // Perform data adaption to the new grid
    virtual void postAdapt() = 0;
  };


  /// \brief Implementation of \ref GridTransferInterface for concrete Grid type.
  template <class Grid>
  class GridTransfer
      : public GridTransferInterface
  {
    using Self = GridTransfer;

  public:
    /// Bind a (mutable) grid to this GridTransfer. Must be called before any xxxAdapt() method.
    // [[ensures: bound()]]
    void bind(Grid& grid)
    {
      grid_ = &grid;
    }

    /// Release the grid from this GridTransfer
    // [[ensures: not bound()]]
    void unbind()
    {
      grid_ = nullptr;
    }

    /// Returns whether this GridTransfer is bound to a grid.
    bool bound() const
    {
      return grid_ != nullptr;
    }

    /// Implements \ref GridTransferInterface::attach
    void attach(DOFVectorInterface* vec) override
    {
      data_.push_back(vec);
    }

    /// Implements \ref GridTransferInterface::detach
    void detach(DOFVectorInterface* vec) override
    {
      auto it = std::find(data_.begin(), data_.end(), vec);
      if (it != data_.end())
        data_.erase(it);
      else
        warning("DOFVector to detach not found");
    }

    /// Implements \ref GridTransferInterface::preAdapt
    // [[expects: bound()]]
    bool preAdapt() override
    {
      assert(bound());
      mightCoarsen_ = grid_->preAdapt();
      for (auto* vec : this->data_)
        vec->preAdapt(mightCoarsen_);
      return mightCoarsen_;
    }

    /// Implements \ref GridTransferInterface::adapt
    // [[expects: bound()]]
    bool adapt() override
    {
      assert(bound());
      refined_ = grid_->adapt();
      return refined_;
    }

    /// Implements \ref GridTransferInterface::postAdapt
    // [[expects: bound()]]
    void postAdapt() override
    {
      assert(bound());
      if (mightCoarsen_ || refined_) {
        for (auto* vec : this->data_)
          vec->postAdapt(refined_);
      }
      grid_->postAdapt();
      changeIndex_++;
    }

    /// Returns the grid change index, see \ref changeIndex.
    unsigned long changeIndex() const
    {
      return changeIndex_;
    }

  private:
    /// A grid this GridTransfer is bound to in \ref bind()
    Grid* grid_ = nullptr;

    /// A list of data containers handled during grid adaption
    std::list<DOFVectorInterface*> data_;

    /// Flag set during \ref preAdapt(), indicating whether any element might be coarsened in \ref adapt()
    bool mightCoarsen_ = false;

    /// Flag set during \ref adapt() indicating that at least one entity was refined
    bool refined_ = false;

    /// This index is incremented every time the grid is changed, e.g. by refinement or coarsening.
    unsigned long changeIndex_ = 0;
  };

  /// @}

} // end namespace AMDiS
