#pragma once

#include <dune/istl/preconditioners.hh>

#include <amdis/CreatorInterface.hpp>

namespace AMDiS
{
  template <class Matrix, class VectorX, class VectorB>
  struct ISTLPreconInterface
  {
    virtual ~ISTLPreconInterface() = default;
    using PreconBase = Dune::Preconditioner<VectorX, VectorB>;
    virtual std::unique_ptr<PreconBase> create(Matrix const& matrix) const = 0;
  };

  template <class> struct Type {};

  // creators for preconditioners, depending on matrix-type
  // ---------------------------------------------------------------------------

  template <class Precon, class Matrix>
  struct ISTLPrecon
      : public ISTLPreconInterface<Matrix, typename Precon::domain_type, typename Precon::range_type>
  {
    using VectorX = typename Precon::domain_type;
    using VectorB = typename Precon::range_type;
    using Super = ISTLPreconInterface<Matrix, VectorX, VectorB>;
    using Self = ISTLPrecon;

    struct Creator : CreatorInterfaceName<Super>
    {
      std::unique_ptr<Super> createWithString(std::string prefix) override
      {
        return std::make_unique<Self>(prefix);
      }
    };

    ISTLPrecon(std::string const& prefix)
    {
      Parameters::get(prefix + "->relaxation", w_);
      Parameters::get(prefix + "->iterations", iter_);
    }

    using PreconBase = Dune::Preconditioner<VectorX, VectorB>;
    std::unique_ptr<PreconBase> create(Matrix const& A) const override
    {
      return createImpl(A, Type<Precon>{});
    }

  private:
    template <class P>
    std::unique_ptr<P> createImpl(Matrix const& A, Type<P>) const
    {
      return std::make_unique<P>(A, iter_, w_);
    }

    std::unique_ptr<Dune::SeqILU<Matrix, VectorX, VectorB>>
    createImpl(Matrix const& A, Type<Dune::SeqILU<Matrix, VectorX, VectorB>>) const
    {
      return std::make_unique<Dune::SeqILU<Matrix, VectorX, VectorB>>(A, iter_, w_);
    }

    std::unique_ptr<Dune::Richardson<VectorX, VectorB>>
    createImpl(Matrix const& /*A*/, Type<Dune::Richardson<VectorX, VectorB>>) const
    {
      return std::make_unique<Dune::Richardson<VectorX, VectorB>>(w_);
    }

    double w_ = 1.0;
    int iter_ = 0;
  };


  /// Adds default creators for linear solvers based on `Dune::BCRSMatrix`.
  /**
   * Adds creators for full-matrix aware solvers.
   * - *cg*: conjugate gradient method, \see Dune::CGSolver
   * - *bcgs*: stabilized bi-conjugate gradient method, \see Dune::BiCGSTABSolver
   * - *minres*: Minimal residul method, \see Dune::MINRESSolver
   * - *gmres*: Generalized minimal residula method, \see Dune::RestartedGMResSolver
   * - *umfpack*: external UMFPACK solver, \see Dune::UMFPack
   **/
  template <class Matrix, class VectorX, class VectorB>
  class DefaultCreators<ISTLPreconInterface<Matrix, VectorX, VectorB>>
  {
    template <class Precon>
    using PreconCreator
      = typename ISTLPrecon<Precon, Matrix>::Creator;

    using Map = CreatorMap<ISTLPreconInterface<Matrix, VectorX, VectorB>>;

  public:
    static void init()
    {
      auto jacobi = new PreconCreator<Dune::SeqJac<Matrix, VectorX, VectorB>>;
      Map::addCreator("diag", jacobi);
      Map::addCreator("jacobi", jacobi);

      auto gs = new PreconCreator<Dune::SeqGS<Matrix, VectorX, VectorB>>;
      Map::addCreator("gs", gs);
      Map::addCreator("gauss_seidel", gs);

      auto sor = new PreconCreator<Dune::SeqSOR<Matrix, VectorX, VectorB>>;
      Map::addCreator("sor", sor);

      auto ssor = new PreconCreator<Dune::SeqSSOR<Matrix, VectorX, VectorB>>;
      Map::addCreator("ssor", ssor);

      auto ilu = new PreconCreator<Dune::SeqILU<Matrix, VectorX, VectorB>>;
      Map::addCreator("ilu", ilu);
      Map::addCreator("ilu0", ilu);

      auto richardson = new PreconCreator<Dune::Richardson<VectorX, VectorB>>;
      Map::addCreator("richardson", richardson);
      Map::addCreator("default", richardson);
    }
  };

} // end namespace AMDiS
