#pragma once

#include <type_traits>

#include <amdis/GridFunctionOperator.hpp>
#include <amdis/utility/LocalBasisCache.hpp>
#include <amdis/Output.hpp>
#include <amdis/common/StaticSize.hpp>

namespace AMDiS
{
  /**
   * \addtogroup operators
   * @{
   **/

  namespace tag
  {
    struct test_partialtrial
    {
      std::size_t comp;
    };

    struct partial_trial
    {
      std::size_t comp;
    };
  }


  /// first-order operator \f$ \langle\psi, c\,\partial_i\phi\rangle \f$
  template <class LC, class GridFct>
  class GridFunctionOperator<tag::test_partialtrial, LC, GridFct>
      : public GridFunctionOperatorBase<GridFunctionOperator<tag::test_partialtrial, LC, GridFct>, LC, GridFct>
  {
    using Self = GridFunctionOperator;
    using Super = GridFunctionOperatorBase<Self, LC, GridFct>;

    static_assert( Size_v<typename GridFct::Range> == 1, "Expression must be of scalar type." );

  public:
    GridFunctionOperator(tag::test_partialtrial tag, GridFct const& expr)
      : Super(expr, 1)
      , comp_(tag.comp)
    {}

    template <class CG, class RN, class CN, class Mat>
    void getElementMatrix(CG const& contextGeo, RN const& rowNode, CN const& colNode, Mat& elementMatrix)
    {
      static_assert(RN::isLeaf && CN::isLeaf,
        "Operator can be applied to Leaf-Nodes only.");

      auto const& quad = this->getQuadratureRule(contextGeo.type(), rowNode, colNode);
      auto const& rowLocalFE = rowNode.finiteElement();
      auto const& colLocalFE = colNode.finiteElement();
      std::size_t rowSize = rowLocalFE.size();
      std::size_t colSize = colLocalFE.size();

      using RangeFieldType = typename NodeQuadCache<CN>::Traits::RangeFieldType;

      NodeQuadCache<RN> rowCache(rowLocalFE.localBasis());
      NodeQuadCache<CN> colCache(colLocalFE.localBasis());

      auto const& shapeValuesCache = rowCache.evaluateFunctionAtQP(contextGeo, quad);
      auto const& shapeGradientsCache = colCache.evaluateJacobianAtQP(contextGeo, quad);
      for (std::size_t iq = 0; iq < quad.size(); ++iq) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = contextGeo.local(quad[iq].position());

        // The transposed inverse Jacobian of the map from the reference element to the element
        const auto jacobian = contextGeo.geometry().jacobianInverseTransposed(local);
        assert(jacobian.M() == CG::dim);

        // The multiplicative factor in the integral transformation formula
        const auto factor = contextGeo.integrationElement(quad[iq].position()) * quad[iq].weight();
        const auto c = Super::coefficient(local);

        // the values of the shape functions on the reference element at the quadrature point
        auto const& shapeValues = shapeValuesCache[iq];

        // The gradients of the shape functions on the reference element
        auto const& shapeGradients = shapeGradientsCache[iq];

        // Compute the shape function gradients on the real element
        thread_local std::vector<RangeFieldType> colPartial;
        colPartial.resize(shapeGradients.size());
        for (std::size_t i = 0; i < colPartial.size(); ++i) {
          colPartial[i] = jacobian[comp_][0] * shapeGradients[i][0][0];
          for (std::size_t j = 1; j < jacobian.M(); ++j)
            colPartial[i] += jacobian[comp_][j] * shapeGradients[i][0][j];
        }

        for (std::size_t j = 0; j < colSize; ++j) {
          const auto local_j = colNode.localIndex(j);
          const auto value = factor * (c * colPartial[j]);
          for (std::size_t i = 0; i < rowSize; ++i) {
            const auto local_i = rowNode.localIndex(i);
            elementMatrix[local_i][local_j] += value * shapeValues[i];
          }
        }
      }
    }

  private:
    std::size_t comp_;
  };


  /// Create a first-order term with derivative on trial-function
  template <class Expr, class... QuadratureArgs>
  auto fot(Expr&& expr, tag::partial_trial t, QuadratureArgs&&... args)
  {
    return makeOperator(tag::test_partialtrial{t.comp}, FWD(expr), FWD(args)...);
  }

  /** @} **/

} // end namespace AMDiS
